import json

from ....errors import (
    ApiError,
    QueryRunCancelledError,
    QueryRunExecutionError,
    QueryRunTimeoutError,
)
from ....errors.api_error import error_codes
from ....integrations import CompassQueryIntegration
from ....integrations.query_integration.defaults import DEFAULTS
from ....models import Query, QueryStatus
from ....models.compass.core.rpc_error import RpcError
from ....rpc import RPC
from ...utils.mock_data.create_query_run import create_query_run_response
from ...utils.mock_data.get_query_results import get_query_results_response
from ...utils.mock_data.get_query_run import get_query_run_response
from ...utils.mock_data.get_sql_statement import get_sql_statement_response

SDK_VERSION = "1.0.2"
SDK_PACKAGE = "python"


def get_rpc():
    return RPC("https://test.com", "api_key")


def test_query_defaults():
    qi = CompassQueryIntegration(get_rpc())

    # Test that the defaults are semi-overridden
    q = Query(sql="", ttl_minutes=5, page_number=5, page_size=10, sdk_package=SDK_PACKAGE, sdk_version=SDK_VERSION)  # type: ignore
    next_q = qi._set_query_defaults(q)

    assert next_q.page_number == 5
    assert next_q.page_size == 10
    assert next_q.ttl_minutes == 5
    assert next_q.sdk_package == SDK_PACKAGE
    assert next_q.sdk_version == SDK_VERSION
    assert next_q.cached == DEFAULTS.cached
    assert next_q.timeout_minutes == DEFAULTS.timeout_minutes

    # Test that the defaults are not overridden
    q = Query(sql="", sdk_package=SDK_PACKAGE, sdk_version=SDK_VERSION)  # type: ignore
    next_q = qi._set_query_defaults(q)

    assert next_q.page_number == DEFAULTS.page_number
    assert next_q.page_size == DEFAULTS.page_size
    assert next_q.ttl_minutes == DEFAULTS.ttl_minutes
    assert next_q.cached == DEFAULTS.cached
    assert next_q.timeout_minutes == DEFAULTS.timeout_minutes
    assert next_q.sdk_package == SDK_PACKAGE
    assert next_q.sdk_version == SDK_VERSION


def test_query_success(requests_mock):
    rpc = get_rpc()
    qi = CompassQueryIntegration(rpc)

    page_number = 1
    page_size = 10

    requests_mock.reset()
    result = requests_mock.post(
        rpc.url,
        text=json.dumps(
            get_query_run_response(
                status=QueryStatus.Success,
            )
        ),
        status_code=200,
        reason="OK",
    )

    result = qi._get_query_run_loop(
        "test_query_id",
        page_number=page_number,
        page_size=page_size,
        attempts=0,
        timeout_minutes=1,
        retry_interval_seconds=0.0001,
    )
    assert result is not None
    assert result.state == QueryStatus.Success


def test_query_failed(requests_mock):
    rpc = get_rpc()
    qi = CompassQueryIntegration(rpc)

    page_number = 1
    page_size = 10

    # Query Execution Error
    requests_mock.reset()
    result = requests_mock.post(
        rpc.url,
        text=json.dumps(
            get_query_run_response(
                status=QueryStatus.Failed,
            )
        ),
        status_code=200,
        reason="OK",
    )

    try:
        result = qi._get_query_run_loop(
            "test_query_id",
            page_number=page_number,
            page_size=page_size,
            attempts=0,
            timeout_minutes=1,
            retry_interval_seconds=0.0001,
        )
    except Exception as e:
        assert type(e) == QueryRunExecutionError


def test_query_cancelled(requests_mock):
    rpc = get_rpc()
    qi = CompassQueryIntegration(rpc)

    page_number = 1
    page_size = 10

    # Query Execution Error
    requests_mock.reset()
    result = requests_mock.post(
        rpc.url,
        text=json.dumps(
            get_query_run_response(
                status=QueryStatus.Canceled,
            )
        ),
        status_code=200,
        reason="OK",
    )

    try:
        qi._get_query_run_loop(
            "test_query_id",
            page_number=page_number,
            page_size=page_size,
            attempts=0,
            timeout_minutes=1,
            retry_interval_seconds=0.0001,
        )
    except Exception as e:
        assert type(e) == QueryRunCancelledError


def test_query_timeout(requests_mock):
    rpc = get_rpc()
    qi = CompassQueryIntegration(rpc)

    page_number = 1
    page_size = 10

    requests_mock.reset()
    result = requests_mock.post(
        rpc.url,
        text=json.dumps(
            get_query_run_response(
                status=QueryStatus.Running,
            )
        ),
        status_code=200,
        reason="OK",
    )

    try:
        result = qi._get_query_run_loop(
            "test_query_id",
            page_number=page_number,
            page_size=page_size,
            attempts=0,
            timeout_minutes=0.1,
            retry_interval_seconds=0.0001,
        )
    except Exception as e:
        assert type(e) == QueryRunTimeoutError


def test_api_error_codes(requests_mock):
    rpc = get_rpc()
    qi = CompassQueryIntegration(rpc)

    page_number = 1
    page_size = 10

    for key, value in error_codes.items():

        requests_mock.reset()
        result = requests_mock.post(
            rpc.url,
            text=json.dumps(
                get_query_run_response(
                    status=QueryStatus.Running,
                    error=RpcError(code=value, message="", data={}),
                )
            ),
            status_code=200,
            reason="OK",
        )

        try:
            qi._get_query_run_loop(
                f"test_query_{key}",
                page_number=page_number,
                page_size=page_size,
                attempts=0,
                timeout_minutes=0.1,
                retry_interval_seconds=0.0001,
            )
        except Exception as e:
            assert type(e) == ApiError


def test_get_private_query_results(requests_mock):
    rpc = get_rpc()
    qi = CompassQueryIntegration(rpc)

    page_number = 1
    page_size = 10

    # Query Result Success
    requests_mock.reset()
    result = requests_mock.post(
        rpc.url,
        text=json.dumps(
            get_query_results_response(
                status=QueryStatus.Success,
            )
        ),
        status_code=200,
        reason="ok",
    )

    result = qi._get_query_results("test_query_id")
    assert result is not None

    # Query Result Error
    requests_mock.reset()
    result = requests_mock.post(
        rpc.url,
        text=json.dumps(
            get_query_results_response(
                status=QueryStatus.Success,
                error=RpcError(
                    code=-53000,
                    message="query results not found",
                    data={},
                ),
            )
        ),
        status_code=200,
        reason="ok",
    )

    try:
        result = qi._get_query_results("test_query_id")
    except Exception as e:
        assert type(e) == ApiError


def test_cancel_query_run(requests_mock):
    rpc = get_rpc()
    qi = CompassQueryIntegration(rpc)

    requests_mock.reset()
    result = requests_mock.post(
        rpc.url,
        text=json.dumps(
            get_query_run_response(
                status=QueryStatus.Success,
            )
        ),
        status_code=200,
        reason="ok",
    )

    result = qi.cancel_query_run("cancel_query_run_id")
    assert result is not None


def test_get_query_results(requests_mock):
    rpc = get_rpc()
    qi = CompassQueryIntegration(rpc)

    requests_mock.reset()
    result = requests_mock.post(
        rpc.url,
        text=json.dumps(
            get_query_results_response(
                status=QueryStatus.Success,
            )
        ),
        status_code=200,
        reason="ok",
    )

    result = qi.get_query_results("get_query_results_id")
    assert result is not None


def test_get_query_run(requests_mock):
    rpc = get_rpc()
    qi = CompassQueryIntegration(rpc)

    requests_mock.reset()
    result = requests_mock.post(
        rpc.url,
        text=json.dumps(
            get_query_run_response(
                status=QueryStatus.Success,
            )
        ),
        status_code=200,
        reason="ok",
    )

    result = qi.get_query_run("get_query_run_id")
    assert result is not None


def test_get_sql_statement(requests_mock):
    rpc = get_rpc()
    qi = CompassQueryIntegration(rpc)

    sql_id = "get_sql_statement_id"
    requests_mock.reset()
    result = requests_mock.post(
        rpc.url,
        text=json.dumps(
            get_sql_statement_response(
                id=sql_id,
            )
        ),
        status_code=200,
        reason="ok",
    )

    result = qi.get_sql_statement(sql_id)
    assert result.id == sql_id



from typing import Dict, List, Optional, Union

from pydantic import BaseModel

from .core.rpc_request import RpcRequest
from .core.rpc_response import RpcResponse
from .core.sql_statement import SqlStatement


# Request
class GetSqlStatementParams(BaseModel):
    sqlStatementId: str


class GetSqlStatementRequest(RpcRequest):
    method: str = "getSqlStatement"
    params: List[GetSqlStatementParams]


# Response
class GetSqlStatemetnResult(BaseModel):
    sqlStatement: SqlStatement


class GetSqlStatementResponse(RpcResponse):
    result: Union[GetSqlStatemetnResult, None]



from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from .tags import Tags


class QueryRequest(BaseModel):
    id: str
    sqlStatementId: str
    userId: str
    tags: Tags
    maxAgeMinutes: int
    resultTTLHours: int
    userSkipCache: bool
    triggeredQueryRun: bool
    queryRunId: str
    createdAt: str
    updatedAt: str



from typing import Optional, Union

from .base_error import BaseError


class QueryRunRateLimitError(BaseError):
    """
    Base class for all QueryRunRateLimitError errors.
    """

    def __init__(self):
        self.message = "QUERY_RUN_RATE_LIMIT_ERROR: you have exceeded the rate limit for creating/running new queries"
        super().__init__(self.message)


class QueryRunTimeoutError(BaseError):
    """
    Base class for all QueryRunTimeoutError errors.
    """

    def __init__(self, timeoutMinutes: Union[int, float, None] = None):
        if timeoutMinutes is None:
            self.message = f"QUERY_RUN_TIMEOUT_ERROR: your query has timed out."
        else:
            self.message = f"QUERY_RUN_TIMEOUT_ERROR: your query has timed out after {timeoutMinutes} minutes."
        super().__init__(self.message)


class QueryRunExecutionError(BaseError):
    """
    Base class for all QueryRunExecutionError errors.
    """

    def __init__(
        self,
        error_name: Optional[str] = None,
        error_message: Optional[str] = None,
        error_data: Optional[str] = None,
    ):
        self.message = f"QUERY_RUN_EXECUTION_ERROR: an error has occured while executing your query. errorName={error_name}, errorMessage={error_message}, errorData={error_data}"
        super().__init__(self.message)


class QueryRunCancelledError(BaseError):
    """
    Base class for all QueryRunCancelledError errors.
    """

    def __init__(
        self,
        error_name: Optional[str] = None,
        error_message: Optional[str] = None,
        error_data: Optional[str] = None,
    ):
        self.message = f"QUERY_RUN_CANCELLED_ERROR: your query has been cancelled. errorName={error_name}, errorMessage={error_message}, errorData={error_data}"
        super().__init__(self.message)


class QueryRunInvalidStateToCancel(BaseError):
    """
    Base class for all QueryRunInvalidStateToCancel errors.
    """

    def __init__(
        self,
        msg: Optional[str] = None,
    ):
        self.message = msg
        super().__init__(self.message)



from typing import Any, List, Union

from pydantic import BaseModel, Field

from .query_run_stats import QueryRunStats


class QueryResultSet(BaseModel):
    query_id: Union[str, None] = Field(None, description="The server id of the query")

    status: str = Field(
        False, description="The status of the query (`PENDING`, `FINISHED`, `ERROR`)"
    )
    columns: Union[List[str], None] = Field(
        None, description="The names of the columns in the result set"
    )
    column_types: Union[List[str], None] = Field(
        None, description="The type of the columns in the result set"
    )
    rows: Union[List[Any], None] = Field(None, description="The results of the query")
    run_stats: Union[QueryRunStats, None] = Field(
        None,
        description="Summary stats on the query run (i.e. the number of rows returned, the elapsed time, etc)",
    )
    records: Union[List[Any], None] = Field(
        None, description="The results of the query transformed as an array of objects"
    )
    error: Any



from datetime import datetime
from typing import Dict, List, Optional

from pydantic import BaseModel


class Tags(BaseModel):
    sdk_package: Optional[str]
    sdk_version: Optional[str]
    sdk_language: Optional[str]



from typing import Any, Dict, List, Union

from pydantic import BaseModel

from .core.query_request import QueryRequest
from .core.query_run import QueryRun
from .core.rpc_request import RpcRequest
from .core.rpc_response import RpcResponse
from .core.sql_statement import SqlStatement
from .core.tags import Tags


# Request
class CreateQueryRunRpcParams(BaseModel):
    resultTTLHours: int
    maxAgeMinutes: int
    sql: str
    tags: Tags
    dataSource: str
    dataProvider: str


class CreateQueryRunRpcRequest(RpcRequest):
    method: str = "createQueryRun"
    params: List[CreateQueryRunRpcParams]


# Response
class CreateQueryRunRpcResult(BaseModel):
    queryRequest: QueryRequest
    queryRun: QueryRun
    sqlStatement: SqlStatement


class CreateQueryRunRpcResponse(RpcResponse):
    result: Union[CreateQueryRunRpcResult, None]



from ...models.sleep_config import SleepConfig
from ...utils.sleep import (
    exp_backoff,
    get_elapsed_exp_seconds,
    get_elapsed_linear_seconds,
    get_exp_backoff_seconds,
    get_linear_backoff_seconds,
    linear_backoff,
    sec_to_ms,
)


def test_sec_to_ms():
    assert sec_to_ms(1) == 1000


def test_get_exp_backoff_seconds():
    assert get_exp_backoff_seconds(0) == 1
    assert get_exp_backoff_seconds(2) == 4


def test_get_linear_backoff_seconds():
    assert get_linear_backoff_seconds(0, 0) == 0
    assert get_linear_backoff_seconds(0, 1) == 1
    assert get_linear_backoff_seconds(1, 1) == 2
    assert get_linear_backoff_seconds(1, 2) == 4


def test_get_elapsed_exp_seconds():
    sc = SleepConfig(attempts=1, interval_seconds=5, timeout_minutes=5)
    assert get_elapsed_exp_seconds(sc) == 1

    sc = SleepConfig(attempts=2, interval_seconds=5, timeout_minutes=5)
    assert get_elapsed_exp_seconds(sc) == 3

    sc = SleepConfig(attempts=3, interval_seconds=5, timeout_minutes=5)
    assert get_elapsed_exp_seconds(sc) == 7


def test_exp_backoff():
    sc = SleepConfig(attempts=1, interval_seconds=0.1, timeout_minutes=5)
    should_continue = exp_backoff(sc)
    assert should_continue is True

    sc = SleepConfig(attempts=1, interval_seconds=0.1, timeout_minutes=0.0001)
    should_continue = exp_backoff(sc)
    assert should_continue is False


def test_get_elapsed_linear_seconds():
    sc = SleepConfig(attempts=0, interval_seconds=5, timeout_minutes=5)
    assert get_elapsed_linear_seconds(sc) == 0

    sc = SleepConfig(attempts=1, interval_seconds=5, timeout_minutes=5)
    assert get_elapsed_linear_seconds(sc) == 5

    sc = SleepConfig(attempts=2, interval_seconds=5, timeout_minutes=5)
    assert get_elapsed_linear_seconds(sc) == 15

    sc = SleepConfig(attempts=3, interval_seconds=5, timeout_minutes=5)
    assert get_elapsed_linear_seconds(sc) == 30


def test_linear_backoff():
    sc = SleepConfig(attempts=1, interval_seconds=0.1, timeout_minutes=5)
    should_continue = linear_backoff(sc)
    assert should_continue is True

    sc = SleepConfig(attempts=1, interval_seconds=0.1, timeout_minutes=0.0001)
    should_continue = linear_backoff(sc)
    assert should_continue is False



from datetime import datetime

from pydantic import BaseModel, Field


class QueryRunStats(BaseModel):
    started_at: datetime = Field(None, description="The start time of the query run.")
    ended_at: datetime = Field(None, description="The end time of the query run.")
    query_exec_started_at: datetime = Field(
        None, description="The start time of query execution."
    )
    query_exec_ended_at: datetime = Field(
        None, description="The end time of query execution."
    )
    streaming_started_at: datetime = Field(
        None, description="The start time of streaming query results."
    )
    streaming_ended_at: datetime = Field(
        None, description="The end time of streaming query results."
    )
    elapsed_seconds: int = Field(
        None,
        description="The number of seconds elapsed between the start and end times.",
    )
    queued_seconds: int = Field(
        None,
        description="The number of seconds elapsed between when the query was created and when execution on the data source began.",
    )
    streaming_seconds: int = Field(
        None,
        description="The number of seconds elapsed between when the query execution completed and results were fully streamed to Flipside's servers.",
    )
    query_exec_seconds: int = Field(
        None,
        description="The number of seconds elapsed between when the query execution started and when it completed on the data source.",
    )
    record_count: int = Field(
        None, description="The number of records returned by the query."
    )
    bytes: int = Field(None, description="The number of bytes returned by the query.")



from typing import List, Union

from pydantic import BaseModel

from .core.query_run import QueryRun
from .core.rpc_request import RpcRequest
from .core.rpc_response import RpcResponse


# Request
class CancelQueryRunRpcRequestParams(BaseModel):
    queryRunId: str


class CancelQueryRunRpcRequest(RpcRequest):
    method: str = "cancelQueryRun"
    params: List[CancelQueryRunRpcRequestParams]


# Response
class CancelQueryRunRpcResult(BaseModel):
    queryRun: QueryRun


class CancelQueryRunRpcResponse(RpcResponse):
    result: Union[CancelQueryRunRpcResult, None]



class ApiError(Exception):
    def __init__(self, name, code, message):
        super().__init__(f"{name}: message={message}, code={code}")


error_codes = {
    "MethodValidationError": -32000,
    "QueryRunNotFound": -32099,
    "SqlStatementNotFound": -32100,
    "TemporalError": -32150,
    "QueryRunNotFinished": -32151,
    "ResultTransformError": -32152,
    "ResultFormatNotSupported": -32153,
    "RowCountCouldNotBeComputed": -32154,
    "QueryResultColumnMetadataMissing": -32155,
    "InvalidSortColumn": -32156,
    "ColumnSummaryQueryFailed": -32157,
    "QueryResultColumnMetadataMissingColumnName": -32158,
    "QueryResultColumnMetadataMissingColumnType": -32159,
    "NoQueryRunsFoundinQueryText": -32160,
    "DuckDBError": -32161,
    "RefreshableQueryNotFound": -32162,
    "AuthorizationError": -32163,
    "DataSourceNotFound": -32164,
    "QueryRunInvalidStateToCancel": -32165,
    "DataProviderAlreadyExists": -32166,
    "DataProviderNotFound": -32167,
    "DataSourceAlreadyExists": -32168,
    "AdminOnly": -32169,
    "RequestedPageSizeTooLarge": -32170,
    "MaxConcurrentQueries": -32171,
}


def get_exception_by_error_code(error_code=None, message=None):
    if not error_code or not message:
        return ApiError("UnknownAPIError", error_code, message)

    error_name = None
    for key, value in error_codes.items():
        if value == error_code:
            error_name = key
            break

    if error_name is None:
        return ApiError("UnknownAPIError", error_code, message)

    return ApiError(error_name, error_code, message)



from datetime import datetime

from ....integrations.query_integration.query_result_set_builder import (
    QueryResultSetBuilder,
)
from ....models.compass.get_query_run import GetQueryRunRpcResponse
from ....models.compass.get_query_run_results import GetQueryRunResultsRpcResponse
from ....models.query_status import QueryStatus
from ...utils.mock_data.get_query_results import get_query_results_response
from ...utils.mock_data.get_query_run import get_query_run_response


def getQueryResultSetData(status: str) -> dict:
    query_run_resp = GetQueryRunRpcResponse(**get_query_run_response(status=status))
    query_result_resp = GetQueryRunResultsRpcResponse(
        **get_query_results_response(status=status)
    )
    return {
        "query_run": query_run_resp.result.queryRun if query_run_resp.result else None,
        "query_result": query_result_resp.result if query_result_resp.result else None,
    }


def test_run_stats():
    qr = QueryResultSetBuilder(**getQueryResultSetData(QueryStatus.Success))

    # Start/end are datetime objects?
    assert type(qr.run_stats.started_at) == datetime
    assert type(qr.run_stats.ended_at) == datetime

    # Elapsed seconds
    assert qr.run_stats.elapsed_seconds == 51

    # Streaming seconds
    assert qr.run_stats.streaming_seconds == 45

    # Exec Seconds
    assert qr.run_stats.query_exec_seconds == 5

    # Record count
    assert qr.run_stats.record_count == 10000


def test_records():
    qr = QueryResultSetBuilder(**getQueryResultSetData(QueryStatus.Success))

    # Records Length Matches Row Length?
    assert qr.records is not None
    assert qr.rows is not None
    assert qr.columns is not None
    assert len(qr.records) == len(qr.rows)

    # Column Length Matches Records Key Length
    for record in qr.records:
        assert record is not None

        assert len(record.keys()) == len(qr.columns)

    # Columns = Record Keys
    for record in qr.records:
        for column in qr.columns:
            assert column in record.keys()

    # Record values match row values?
    for record, row in zip(qr.records, qr.rows):
        for column, value in zip(qr.columns, row):
            assert record[column] == value


def test_status():
    # Status is ready?
    qr = QueryResultSetBuilder(**getQueryResultSetData(QueryStatus.Ready))
    assert qr.status == QueryStatus.Ready

    # Status is running?
    qr = QueryResultSetBuilder(**getQueryResultSetData(QueryStatus.Running))
    assert qr.status == QueryStatus.Running

    # Status is streaming?
    qr = QueryResultSetBuilder(**getQueryResultSetData(QueryStatus.StreamingResults))
    assert qr.status == QueryStatus.StreamingResults

    # Status is cancelled?
    qr = QueryResultSetBuilder(**getQueryResultSetData(QueryStatus.Canceled))
    assert qr.status == QueryStatus.Canceled

    # Status is finished?
    qr = QueryResultSetBuilder(**getQueryResultSetData(QueryStatus.Success))
    assert qr.status == QueryStatus.Success

    # Status is failed?
    qr = QueryResultSetBuilder(**getQueryResultSetData(QueryStatus.Failed))
    assert qr.status == QueryStatus.Failed


def test_query_id():
    # Query ID is set?
    qr = QueryResultSetBuilder(**getQueryResultSetData(QueryStatus.Success))
    assert qr.query_id is not None

    # Query ID is test
    qr = QueryResultSetBuilder(**getQueryResultSetData(QueryStatus.Success))
    assert qr.query_id == "clg44olzq00cbn60tasvob5l2"



from ...models.query_status import (
    QueryStatus,
    QueryStatusCanceled,
    QueryStatusFailed,
    QueryStatusReady,
    QueryStatusRunning,
    QueryStatusStreamingResults,
    QueryStatusSuccess,
)


def test_query_status():
    assert QueryStatusReady == "QUERY_STATE_READY"
    assert QueryStatusRunning == "QUERY_STATE_RUNNING"
    assert QueryStatusStreamingResults == "QUERY_STATE_STREAMING_RESULTS"
    assert QueryStatusSuccess == "QUERY_STATE_SUCCESS"
    assert QueryStatusFailed == "QUERY_STATE_FAILED"
    assert QueryStatusCanceled == "QUERY_STATE_CANCELED"

    assert QueryStatus.Ready == "QUERY_STATE_READY"
    assert QueryStatus.Running == "QUERY_STATE_RUNNING"
    assert QueryStatus.StreamingResults == "QUERY_STATE_STREAMING_RESULTS"
    assert QueryStatus.Success == "QUERY_STATE_SUCCESS"
    assert QueryStatus.Failed == "QUERY_STATE_FAILED"
    assert QueryStatus.Canceled == "QUERY_STATE_CANCELED"



from typing import Union

from .base_error import BaseError


class SDKError(BaseError):
    """
    Base class for all SDK errors.
    """

    def __init__(self, message: Union[str, None]):
        self.message = message
        super().__init__(self.message)



import time
from typing import Union

from ..models.sleep_config import SleepConfig


def sec_to_ms(sec: int):
    return sec * 1000


def get_exp_backoff_seconds(attempts: Union[int, float]):
    return 2**attempts


def get_linear_backoff_seconds(
    attempts: Union[int, float], interval_seconds: Union[int, float]
):
    return (attempts + 1) * interval_seconds


def get_elapsed_exp_seconds(config: SleepConfig):
    if not config.interval_seconds:
        raise Exception("intervalSeconds is required for `get_elapsed_exp_seconds`")

    elapsed_seconds = 0
    for i in range(config.attempts):
        elapsed_seconds += get_exp_backoff_seconds(i)

    return elapsed_seconds


def exp_backoff(config: SleepConfig):
    if not config.interval_seconds:
        raise Exception("intervalSeconds is required for `exp_backoff`")

    elapsed_seconds = get_elapsed_exp_seconds(config)
    should_continue = True
    if elapsed_seconds / 60 > config.timeout_minutes:
        should_continue = False
        return should_continue

    time.sleep(get_exp_backoff_seconds(config.attempts * config.interval_seconds))

    return should_continue


def get_elapsed_linear_seconds(config: SleepConfig):
    if not config.interval_seconds:
        raise Exception("intervalSeconds is required for `get_elapsed_linear_seconds`")

    elapsed_seconds = 0
    for i in range(config.attempts):
        elapsed_seconds += get_linear_backoff_seconds(i, config.interval_seconds)

    return elapsed_seconds


def linear_backoff(config: SleepConfig):
    if not config.interval_seconds:
        raise Exception("intervalSeconds is required for `linear_backoff`")

    elapsed_seconds = get_elapsed_linear_seconds(config)
    should_continue = True
    if elapsed_seconds / 60 > config.timeout_minutes:
        should_continue = False
        return should_continue

    time.sleep(get_linear_backoff_seconds(config.attempts, config.interval_seconds))

    return should_continue



from datetime import datetime
from typing import List, Union

from ...models import QueryResultSet, QueryRunStats
from ...models.compass.core.query_run import QueryRun
from ...models.compass.get_query_run_results import GetQueryRunResultsRpcResult


class QueryResultSetBuilder(object):
    def __init__(self, query_run: QueryRun, query_result: GetQueryRunResultsRpcResult):
        self.query_id = query_run.id
        self.status = query_run.state
        self.columns = query_result.columnNames
        self.column_types = query_result.columnTypes
        self.rows = query_result.rows
        self.tags = query_run.tags
        self.sqlStatementId = query_run.sqlStatementId
        self.dataSourceId = query_run.dataSourceId
        self.errorName = query_run.errorName
        self.errorMessage = query_run.errorMessage
        self.errorData = query_run.errorData
        self.dataSourceQueryId = query_run.dataSourceQueryId
        self.dataSourceSessionId = query_run.dataSourceSessionId
        self.path = query_run.path

        self.run_stats = self.compute_run_stats(query_run)
        self.records = self.create_records()

    def build(self) -> QueryResultSet:
        return QueryResultSet(
            query_id=self.query_id,
            status=self.status,
            columns=self.columns,
            column_types=self.column_types,
            rows=self.rows,
            run_stats=self.run_stats,
            records=self.records,
            error=None,
        )

    def compute_run_stats(self, query_run: QueryRun) -> QueryRunStats:
        if (
            not query_run.startedAt
            or not query_run.endedAt
            or not query_run.createdAt
            or not query_run.queryStreamingEndedAt
            or not query_run.queryRunningEndedAt
        ):
            raise Exception("Query has no data ")

        created_at = datetime.strptime(query_run.createdAt, "%Y-%m-%dT%H:%M:%S.%fZ")
        start_time = datetime.strptime(query_run.startedAt, "%Y-%m-%dT%H:%M:%S.%fZ")
        end_time = datetime.strptime(query_run.endedAt, "%Y-%m-%dT%H:%M:%S.%fZ")
        streaming_end_time = datetime.strptime(
            query_run.queryStreamingEndedAt, "%Y-%m-%dT%H:%M:%S.%fZ"
        )
        query_exec_end_at = datetime.strptime(
            query_run.queryRunningEndedAt, "%Y-%m-%dT%H:%M:%S.%fZ"
        )

        return QueryRunStats(
            started_at=start_time,
            ended_at=end_time,
            streaming_started_at=query_exec_end_at,
            streaming_ended_at=streaming_end_time,
            query_exec_started_at=start_time,
            query_exec_ended_at=query_exec_end_at,
            elapsed_seconds=(end_time - start_time).seconds,
            queued_seconds=(start_time - created_at).seconds,
            streaming_seconds=(streaming_end_time - query_exec_end_at).seconds,
            query_exec_seconds=(query_exec_end_at - start_time).seconds,
            record_count=query_run.rowCount if query_run.rowCount else 0,
            bytes=query_run.totalSize if query_run.totalSize else 0,
        )

    def create_records(self) -> Union[List[dict], None]:
        if not self.rows:
            return None

        column_labels = self.columns
        if not column_labels:
            return None

        records: List[dict] = []
        for row in self.rows:
            if not row:
                continue

            record = {}
            for i, col in enumerate(column_labels):
                record[col.lower()] = row[i]

            records.append(record)

        return records



from typing import Dict, List, Optional, Union

from pydantic import BaseModel

from .core.query_run import QueryRun
from .core.rpc_request import RpcRequest
from .core.rpc_response import RpcResponse
from .core.tags import Tags


# Request
class GetQueryRunRpcRequestParams(BaseModel):
    queryRunId: str


class GetQueryRunRpcRequest(RpcRequest):
    method: str = "getQueryRun"
    params: List[GetQueryRunRpcRequestParams]


# Response
class GetQueryRunRpcResult(BaseModel):
    queryRun: QueryRun
    redirectedToQueryRun: Optional[QueryRun]


class GetQueryRunRpcResponse(RpcResponse):
    result: Union[GetQueryRunRpcResult, None]



from typing import List, Optional, Union

from .integrations.query_integration.compass_query_integration import (
    CompassQueryIntegration,
)
from .models import Filter, Query, SortBy
from .models.compass.core.query_run import QueryRun
from .models.compass.core.sql_statement import SqlStatement
from .models.query_result_set import QueryResultSet
from .rpc import RPC

API_BASE_URL = "https://api-v2.flipsidecrypto.xyz"

DEFAULT_DATA_SOURCE = "snowflake-default"
DEFAULT_DATA_PROVIDER = "flipside"

SDK_VERSION = "2.0.2"
SDK_PACKAGE = "python"


class Flipside(object):
    def __init__(self, api_key: str, api_base_url: str = API_BASE_URL):
        if "api." in api_base_url.lower():
            api_base_url = API_BASE_URL
        self.rpc = RPC(api_base_url, api_key)
        self.query_integration = CompassQueryIntegration(self.rpc)

    def query(
        self,
        sql,
        ttl_minutes=30,
        max_age_minutes=0,
        cached=True,
        timeout_minutes=20,
        retry_interval_seconds=1,
        page_size=100000,
        page_number=1,
        data_source=DEFAULT_DATA_SOURCE,
        data_provider=DEFAULT_DATA_PROVIDER,
    ) -> QueryResultSet:
        query_integration = CompassQueryIntegration(self.rpc)

        return query_integration.run(
            Query(
                sql=sql,
                ttl_minutes=ttl_minutes,
                timeout_minutes=timeout_minutes,
                max_age_minutes=max_age_minutes,
                retry_interval_seconds=retry_interval_seconds,
                page_size=page_size,
                page_number=page_number,
                cached=cached,
                sdk_package=SDK_PACKAGE,
                sdk_version=SDK_VERSION,
                data_source=data_source,
                data_provider=data_provider,
            )
        )

    def get_query_run(self, query_run_id: str) -> QueryRun:
        return self.query_integration.get_query_run(query_run_id)

    def get_query_results(
        self,
        query_run_id: str,
        page_number: int = 1,
        page_size: int = 10000,
        filters: Optional[Union[List[Filter], None]] = [],
        sort_by: Optional[Union[List[SortBy], None]] = [],
    ) -> QueryResultSet:
        return self.query_integration.get_query_results(
            query_run_id,
            page_number=page_number,
            page_size=page_size,
            filters=filters,
            sort_by=sort_by,
        )

    def get_sql_statement(self, sql_statement_id: str) -> SqlStatement:
        return self.query_integration.get_sql_statement(sql_statement_id)

    def cancel_query_run(self, query_run_id: str) -> QueryRun:
        return self.query_integration.cancel_query_run(query_run_id)


class ShroomDK(Flipside):
    pass



from typing import Any, Dict, Optional, Union

from pydantic import BaseModel

from .rpc_error import RpcError


class RpcResponse(BaseModel):
    jsonrpc: str
    id: int
    result: Union[Optional[Dict[str, Any]], None]
    error: Optional[RpcError]



QueryStatusReady = "QUERY_STATE_READY"
QueryStatusRunning = "QUERY_STATE_RUNNING"
QueryStatusStreamingResults = "QUERY_STATE_STREAMING_RESULTS"
QueryStatusSuccess = "QUERY_STATE_SUCCESS"
QueryStatusFailed = "QUERY_STATE_FAILED"
QueryStatusCanceled = "QUERY_STATE_CANCELED"


class QueryStatus(object):
    Ready: str = QueryStatusReady
    Running: str = QueryStatusRunning
    StreamingResults: str = QueryStatusStreamingResults
    Success: str = QueryStatusSuccess
    Failed: str = QueryStatusFailed
    Canceled: str = QueryStatusCanceled



from typing import Dict, List

from pydantic import BaseModel

from .core.page import Page
from .core.page_stats import PageStats
from .core.query_run import QueryRun
from .core.result_format import ResultFormat
from .core.rpc_request import RpcRequest
from .core.rpc_response import RpcResponse


# Request
class QueryResultsRpcParams(BaseModel):
    query: str
    format: ResultFormat
    page: Page


class QueryResultsRpcRequest(RpcRequest):
    method: str = "queryResults"
    params: List[QueryResultsRpcParams]


# Response
class QueryResultsRpcResult(BaseModel):
    columnNames: List[str]
    columnTypes: List[str]
    rows: List[Dict]
    page: PageStats
    sql: str
    format: ResultFormat
    originalQueryRun: QueryRun
    redirectedToQueryRun: QueryRun


class QueryResultsRpcResponse(RpcResponse):
    result: QueryResultsRpcResult



import json
from typing import List, Optional, Union

from ...errors import (
    QueryRunCancelledError,
    QueryRunExecutionError,
    QueryRunTimeoutError,
    SDKError,
)
from ...errors.api_error import get_exception_by_error_code
from ...errors.query_run_errors import QueryRunInvalidStateToCancel
from ...models import Query, QueryDefaults, QueryResultSet, QueryStatus, SleepConfig
from ...models.compass.cancel_query_run import CancelQueryRunRpcRequestParams
from ...models.compass.core.page import Page
from ...models.compass.core.query_run import QueryRun
from ...models.compass.core.result_format import ResultFormat
from ...models.compass.core.sql_statement import SqlStatement
from ...models.compass.core.tags import Tags
from ...models.compass.create_query_run import CreateQueryRunRpcParams
from ...models.compass.get_query_run import GetQueryRunRpcRequestParams
from ...models.compass.get_query_run_results import (
    Filter,
    GetQueryRunResultsRpcParams,
    GetQueryRunResultsRpcResult,
    SortBy,
)
from ...models.compass.get_sql_statement import GetSqlStatementParams
from ...rpc import RPC
from ...utils.sleep import get_elapsed_linear_seconds, linear_backoff
from .defaults import DEFAULTS
from .query_result_set_builder import QueryResultSetBuilder


class CompassQueryIntegration(object):
    def __init__(self, rpc: RPC, defaults: QueryDefaults = DEFAULTS):
        self.rpc = rpc
        self.defaults = defaults

    def run(self, query: Query) -> QueryResultSet:
        query = self._set_query_defaults(query)

        create_query_run_params = CreateQueryRunRpcParams(
            resultTTLHours=int(query.ttl_minutes / 60)
            if query.ttl_minutes
            else DEFAULTS.ttl_minutes,
            sql=query.sql,
            maxAgeMinutes=query.max_age_minutes
            if query.max_age_minutes
            else DEFAULTS.max_age_minutes,
            tags=Tags(
                sdk_language="python",
                sdk_package=query.sdk_package,
                sdk_version=query.sdk_version,
            ),
            dataSource=query.data_source if query.data_source else "snowflake-default",
            dataProvider=query.data_provider if query.data_provider else "flipside",
        )
        created_query = self.rpc.create_query(create_query_run_params)
        if created_query.error:
            raise get_exception_by_error_code(
                error_code=created_query.error.code if created_query.error else None,
                message=created_query.error.message if created_query.error else None,
            )

        if not created_query.result or not created_query.result.queryRun:
            raise SDKError("expected `query_run` from server but got `None`")

        query_run = self._get_query_run_loop(
            created_query.result.queryRun.id,
            page_number=query.page_number,
            page_size=query.page_size,
            timeout_minutes=query.timeout_minutes if query.timeout_minutes else 20,
            retry_interval_seconds=query.retry_interval_seconds
            if query.retry_interval_seconds
            else 1,
        )

        query_result = self._get_query_results(
            query_run.id,
            page_number=query.page_number if query.page_number else 1,
            page_size=query.page_size if query.page_size else 100000,
        )

        return QueryResultSetBuilder(
            query_run=query_run,
            query_result=query_result,
        ).build()

    def get_sql_statement(self, sql_statement_id: str) -> SqlStatement:
        response = self.rpc.get_sql_statement(
            GetSqlStatementParams(**{"sqlStatementId": sql_statement_id})
        )

        if response.error or not response.result:
            raise get_exception_by_error_code(
                error_code=response.error.code if response.error else None,
                message=response.error.message if response.error else None,
            )

        return response.result.sqlStatement

    def get_query_run(self, query_run_id: str) -> QueryRun:
        response = self.rpc.get_query_run(
            GetQueryRunRpcRequestParams(queryRunId=query_run_id)
        )

        if response.error or not response.result:
            raise get_exception_by_error_code(
                error_code=response.error.code if response.error else None,
                message=response.error.message if response.error else None,
            )

        return response.result.queryRun

    def cancel_query_run(self, query_run_id: str) -> QueryRun:
        response = self.rpc.cancel_query_run(
            CancelQueryRunRpcRequestParams(queryRunId=query_run_id)
        )
        if response.error or not response.result:
            raise get_exception_by_error_code(
                error_code=response.error.code if response.error else None,
                message=response.error.message if response.error else None,
            )

        return response.result.queryRun

    def get_query_results(
        self,
        query_run_id: str,
        page_number: int = 1,
        page_size: int = 100000,
        filters: Optional[Union[List[Filter], None]] = [],
        sort_by: Optional[Union[List[SortBy], None]] = [],
    ) -> QueryResultSet:
        query_result = self._get_query_results(
            query_run_id,
            page_number=page_number if page_number else 1,
            page_size=page_size if page_size else 10000,
            filters=filters,
            sort_by=sort_by,
        )

        query_run = (
            query_result.redirectedToQueryRun
            if query_result.redirectedToQueryRun
            else query_result.originalQueryRun
        )

        return QueryResultSetBuilder(
            query_run=query_run,
            query_result=query_result,
        ).build()

    def _get_query_results(
        self,
        query_run_id: str,
        page_number: int = 1,
        page_size: int = 100000,
        filters: Optional[Union[List[Filter], None]] = [],
        sort_by: Optional[Union[List[SortBy], None]] = [],
    ) -> GetQueryRunResultsRpcResult:
        query_results_resp = self.rpc.get_query_result(
            GetQueryRunResultsRpcParams(
                queryRunId=query_run_id,
                format=ResultFormat.csv,
                page=Page(
                    number=page_number,
                    size=page_size,
                ),
                filters=filters,
                sortBy=sort_by,
            )
        )

        if query_results_resp.error:
            raise get_exception_by_error_code(
                error_code=query_results_resp.error.code
                if query_results_resp.error
                else None,
                message=query_results_resp.error.message
                if query_results_resp.error
                else None,
            )

        if not query_results_resp.result:
            raise SDKError("expected `result` from server but got `None`")

        query_results = query_results_resp.result
        return query_results

    def _set_query_defaults(self, query: Query) -> Query:
        query_default_dict = self.defaults.dict()
        query_dict = query.dict()
        query_default_dict.update(
            {k: v for (k, v) in query_dict.items() if v is not None}
        )
        return Query(**query_default_dict)

    def _get_query_run_loop(
        self,
        query_run_id: str,
        page_number: int = 1,
        page_size: int = 100000,
        attempts: int = 0,
        timeout_minutes: Union[int, float] = 20,
        retry_interval_seconds: Union[int, float] = 1.0,
    ) -> QueryRun:
        query_run_rpc_resp = self.rpc.get_query_run(
            GetQueryRunRpcRequestParams(queryRunId=query_run_id)
        )

        if query_run_rpc_resp.error:
            raise get_exception_by_error_code(
                error_code=query_run_rpc_resp.error.code
                if query_run_rpc_resp.error
                else None,
                message=query_run_rpc_resp.error.message
                if query_run_rpc_resp.error
                else None,
            )

        if not query_run_rpc_resp.result:
            raise SDKError(
                "valid status msg returned from server but no data exists in the response"
            )

        query_run = query_run_rpc_resp.result.queryRun
        query_status = query_run.state

        if query_status == QueryStatus.Success:
            return query_run

        if query_status == QueryStatus.Failed:
            if query_run.errorName == "QueryRunTimedOut":
                raise QueryRunTimeoutError()

            raise QueryRunExecutionError(
                error_message=query_run.errorMessage,
                error_name=query_run.errorName,
                error_data=query_run.errorData,
            )

        if query_status == QueryStatus.Canceled:
            raise QueryRunCancelledError(
                error_message=query_run.errorMessage,
                error_name=query_run.errorName,
                error_data=query_run.errorData,
            )

        should_continue = linear_backoff(
            SleepConfig(
                attempts=attempts,
                timeout_minutes=timeout_minutes,
                interval_seconds=retry_interval_seconds,
            )
        )

        if not should_continue:
            elapsed_seconds = get_elapsed_linear_seconds(
                SleepConfig(
                    attempts=attempts,
                    timeout_minutes=timeout_minutes,
                    interval_seconds=retry_interval_seconds,
                )
            )

            raise QueryRunTimeoutError(elapsed_seconds)

        return self._get_query_run_loop(
            query_run_id,
            page_number,
            page_size,
            attempts + 1,
            timeout_minutes,
            retry_interval_seconds,
        )



from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from .column_metadata import ColumnMetadata
from .tags import Tags


class SqlStatement(BaseModel):
    id: str
    statementHash: str
    sql: str
    columnMetadata: Optional[ColumnMetadata]
    userId: str
    tags: Tags
    createdAt: str
    updatedAt: str



from ...models import QueryDefaults

DEFAULTS: QueryDefaults = QueryDefaults(
    ttl_minutes=60,
    cached=True,
    timeout_minutes=20,
    retry_interval_seconds=0.5,
    page_size=100000,
    page_number=1,
    max_age_minutes=5,
)



from typing import Union

from .base_error import BaseError


class ServerError(BaseError):
    """
    Base class for all server errors.
    """

    def __init__(self, status_code: int, message: Union[str, None]):
        self.message = f"unexpected server error occured with status code: {status_code}, msg: {message}"
        super().__init__(self.message)



from .compass import Filter, SortBy  # noqa: F401
from .query import Query  # noqa: F401
from .query_defaults import QueryDefaults  # noqa: F401
from .query_result_set import QueryResultSet  # noqa: F401
from .query_run_stats import QueryRunStats  # noqa: F401
from .query_status import QueryStatus  # noqa: F401
from .sleep_config import SleepConfig  # noqa: F401



import json

from ..errors.server_error import ServerError
from ..models import Query, QueryStatus
from ..models.compass.core.page import Page
from ..models.compass.core.result_format import ResultFormat
from ..models.compass.core.rpc_error import RpcError
from ..models.compass.core.tags import Tags
from ..models.compass.create_query_run import CreateQueryRunRpcParams
from ..models.compass.get_query_run import GetQueryRunRpcRequestParams
from ..models.compass.get_query_run_results import GetQueryRunResultsRpcParams
from ..rpc import RPC
from .utils.mock_data.create_query_run import create_query_run_response
from .utils.mock_data.get_query_results import get_query_results_response
from .utils.mock_data.get_query_run import get_query_run_response

"""
Test Defaults
"""
tags = Tags(sdk_language="python", sdk_package="shroomdk", sdk_version="0.0.1")
SERVICE_URL = "https://rpc.flipsidecrypto.xyz"

"""
CreateQuery Run tests
"""


def test_url(requests_mock):
    rpc = RPC(SERVICE_URL, "api_key")
    assert rpc.url == f"{SERVICE_URL}/json-rpc"


def test_create_query_success(requests_mock):
    rpc = RPC(SERVICE_URL, "api_key")

    result = requests_mock.post(
        rpc.url,
        text=json.dumps(create_query_run_response(status=QueryStatus.Success)),
        status_code=200,
        reason="OK",
    )

    q = CreateQueryRunRpcParams(
        resultTTLHours=1,
        maxAgeMinutes=5,
        sql="SELECT * FROM mytable",
        tags=tags,
        dataSource="snowflake-default",
        dataProvider="flipside",
    )

    result = rpc.create_query(q)

    assert result.result is not None
    assert result.result.queryRun is not None
    assert result.result.queryRequest is not None
    assert result.result.sqlStatement is not None
    assert result.result.queryRun.state == QueryStatus.Success
    assert result.error is None


def test_create_query_server_error(requests_mock):
    rpc = RPC(SERVICE_URL, "api_key")
    result = requests_mock.post(
        rpc.url,
        text="{'bad json', 'data",
        status_code=500,
        reason="OK",
    )

    q = CreateQueryRunRpcParams(
        resultTTLHours=1,
        maxAgeMinutes=5,
        sql="SELECT * FROM mytable",
        tags=tags,
        dataSource="snowflake-default",
        dataProvider="flipside",
    )

    try:
        rpc.create_query(q)
    except Exception as e:
        assert type(e) == ServerError
        return

    assert "ServerError not raised" == False


"""
GetQueryRun tests
"""


def test_get_query_run_success(requests_mock):
    rpc = RPC(SERVICE_URL, "api_key")

    result = requests_mock.post(
        rpc.url,
        text=json.dumps(get_query_run_response(status=QueryStatus.Success)),
        status_code=200,
        reason="OK",
    )

    q = GetQueryRunRpcRequestParams(
        queryRunId="randomid",
    )

    result = rpc.get_query_run(q)

    assert result.result is not None
    assert result.result.queryRun is not None
    assert result.error is None


def test_get_query_run_server_error(requests_mock):
    rpc = RPC(SERVICE_URL, "api_key")
    result = requests_mock.post(
        rpc.url,
        text="{'bad json', 'data",
        status_code=500,
        reason="OK",
    )

    q = GetQueryRunRpcRequestParams(
        queryRunId="randomid",
    )

    try:
        rpc.get_query_run(q)
    except Exception as e:
        assert type(e) == ServerError
        return

    assert "ServerError not raised" == False


"""
GetQueryResult tests
"""


def test_get_query_results_success(requests_mock):
    rpc = RPC(SERVICE_URL, "api_key")

    result = requests_mock.post(
        rpc.url,
        text=json.dumps(get_query_results_response(status=QueryStatus.Success)),
        status_code=200,
        reason="OK",
    )

    q = GetQueryRunResultsRpcParams(
        queryRunId="randomid", format=ResultFormat.csv, page=Page(size=1, number=1)
    )

    result = rpc.get_query_result(q)

    assert result.result is not None
    assert result.result.rows is not None
    assert len(result.result.rows) == q.page.size
    assert result.result.page is not None
    assert result.result.page.currentPageNumber == q.page.number
    assert result.result.page.currentPageSize == q.page.size
    assert result.result.originalQueryRun is not None
    assert result.result.sql is not None
    assert result.result.format is not None
    assert result.result.format == q.format
    assert result.result.columnNames is not None
    assert result.result.columnTypes is not None
    assert len(result.result.columnNames) == len(result.result.columnTypes)
    assert len(result.result.columnNames) == len(result.result.rows[0])
    assert result.error is None


def test_get_query_results_server_error(requests_mock):
    rpc = RPC(SERVICE_URL, "api_key")

    result = requests_mock.post(
        rpc.url,
        text="{'bad json', 'data",
        status_code=200,
        reason="OK",
    )

    q = GetQueryRunResultsRpcParams(
        queryRunId="randomid", format=ResultFormat.csv, page=Page(size=1, number=1)
    )

    try:
        rpc.get_query_result(q)
    except Exception as e:
        assert type(e) == ServerError
        return

    assert "ServerError not raised" == False



from typing import Optional, Union

from pydantic import BaseModel, Field


class Query(BaseModel):
    sql: str = Field(None, description="SQL query to execute")
    ttl_minutes: Optional[int] = Field(
        None, description="The number of minutes to cache the query results"
    )
    max_age_minutes: Optional[int] = Field(
        5, description="The max age of the query results in minutes"
    )
    timeout_minutes: Optional[int] = Field(
        None, description="The number of minutes to timeout the query"
    )
    retry_interval_seconds: Optional[Union[int, float]] = Field(
        1, description="The number of seconds to use between retries"
    )
    cached: Optional[bool] = Field(
        None,
        description="An override on the cache. A value of true will Re-Execute the query.",
    )
    page_size: int = Field(None, description="The number of results to return per page")
    page_number: int = Field(None, description="The page number to return")
    sdk_package: Optional[str] = Field(
        None, description="The SDK package used for the query"
    )
    sdk_version: Optional[str] = Field(
        None, description="The SDK version used for the query"
    )
    data_source: Optional[str] = Field(
        None, description="The data source to execute the query against"
    )
    data_provider: Optional[str] = Field(
        None, description="The owner of the data source"
    )



from setuptools import find_packages, setup

with open("./README.md", "r") as fh:
    long_description = fh.read()


with open("./VERSION") as f:
    version = f.read().strip()


with open("requirements.txt", "r") as fh:
    requirements = fh.readlines()


with open("package_name.txt", "r") as fh:
    package_name = fh.read().strip().lower()

setup(
    install_requires=[req for req in requirements if req[:2] != "# "],
    name=package_name,
    version=version,
    author="dev@flipsidecrypto.com",
    author_email="dev@flipsidecrypto.com",
    description="SDK by Flipside Crypto: Query the most reliable & comprehensive blockchain data in crypto",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/FlipsideCrypto/sdk/python",
    packages=find_packages(exclude=["src"]),  # Add the exclude parameter here
    include_package_data=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",  # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        "Intended Audience :: Developers",  # Define that your audience are developers
        "License :: OSI Approved :: MIT License",  # Again, pick a license
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    dependency_links=[],
    python_requires=">=3.7",
)



from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel

from .core.page import Page
from .core.page_stats import PageStats
from .core.query_run import QueryRun
from .core.result_format import ResultFormat
from .core.rpc_request import RpcRequest
from .core.rpc_response import RpcResponse


# Request
class Filter(BaseModel):
    column: str
    eq: Optional[str] = None
    neq: Optional[str] = None
    gt: Optional[str] = None
    gte: Optional[str] = None
    lt: Optional[str] = None
    lte: Optional[str] = None
    like: Optional[str] = None
    in_: Optional[List[str]] = None
    notIn: Optional[List[str]] = None

    class Config:
        fields = {"in_": "in"}

    def dict(self, *args, **kwargs) -> dict:
        kwargs.setdefault("exclude_none", True)  # Exclude keys with None values
        return super().dict(*args, **kwargs)


class SortBy(BaseModel):
    column: str
    direction: str


class GetQueryRunResultsRpcParams(BaseModel):
    queryRunId: str
    format: ResultFormat
    filters: Optional[Union[List[Filter], None]] = []
    sortBy: Optional[Union[List[SortBy], None]] = []
    page: Page

    def dict(self, *args, **kwargs) -> dict:
        kwargs.setdefault("exclude_none", True)  # Exclude keys with None values
        return super().dict(*args, **kwargs)


class GetQueryRunResultsRpcRequest(RpcRequest):
    method: str = "getQueryRunResults"
    params: List[GetQueryRunResultsRpcParams]

    def dict(self, *args, **kwargs) -> dict:
        kwargs.setdefault("exclude_none", True)  # Exclude keys with None values
        return super().dict(*args, **kwargs)


# Response
class GetQueryRunResultsRpcResult(BaseModel):
    columnNames: Optional[List[str]]
    columnTypes: Optional[List[str]]
    rows: List[Any]
    page: PageStats
    sql: str
    format: ResultFormat
    originalQueryRun: QueryRun
    redirectedToQueryRun: Union[QueryRun, None]


class GetQueryRunResultsRpcResponse(RpcResponse):
    result: Union[GetQueryRunResultsRpcResult, None]



from pydantic import BaseModel, Field


class QueryDefaults(BaseModel):
    ttl_minutes: int = Field(
        None, description="The number of minutes to cache the query results"
    )
    max_age_minutes: int = Field(
        None,
        description="The max age of query results to accept before deciding to run a query again",
    )
    cached: bool = Field(False, description="Whether or not to cache the query results")
    timeout_minutes: int = Field(
        None, description="The number of minutes to timeout the query"
    )
    retry_interval_seconds: float = Field(
        None, description="The number of seconds to wait before retrying the query"
    )
    page_size: int = Field(None, description="The number of results to return per page")
    page_number: int = Field(None, description="The page number to return")



from .api_error import ApiError  # noqa: F401
from .query_run_errors import (  # noqa: F401
    QueryRunCancelledError,
    QueryRunExecutionError,
    QueryRunInvalidStateToCancel,
    QueryRunRateLimitError,
    QueryRunTimeoutError,
)
from .sdk_error import SDKError  # noqa: F401
from .server_error import ServerError  # noqa: F401



import json
from typing import List

import requests
from requests.adapters import HTTPAdapter, Retry

from .errors.api_error import ApiError
from .errors.server_error import ServerError
from .models.compass.cancel_query_run import (
    CancelQueryRunRpcRequest,
    CancelQueryRunRpcRequestParams,
    CancelQueryRunRpcResponse,
)
from .models.compass.create_query_run import (
    CreateQueryRunRpcParams,
    CreateQueryRunRpcRequest,
    CreateQueryRunRpcResponse,
)
from .models.compass.get_query_run import (
    GetQueryRunRpcRequest,
    GetQueryRunRpcRequestParams,
    GetQueryRunRpcResponse,
)
from .models.compass.get_query_run_results import (
    GetQueryRunResultsRpcParams,
    GetQueryRunResultsRpcRequest,
    GetQueryRunResultsRpcResponse,
)
from .models.compass.get_sql_statement import (
    GetSqlStatementParams,
    GetSqlStatementRequest,
    GetSqlStatementResponse,
)


class RPC(object):
    def __init__(
        self,
        base_url: str,
        api_key: str,
        max_retries: int = 10,
        backoff_factor: float = 1,
        status_forcelist: List[int] = [429, 500, 502, 503, 504],
        method_allowlist: List[str] = [
            "HEAD",
            "GET",
            "PUT",
            "POST",
            "DELETE",
            "OPTIONS",
            "TRACE",
        ],
    ):
        self._base_url = base_url
        self._api_key = api_key

        # Session Settings
        self._MAX_RETRIES = max_retries
        self._BACKOFF_FACTOR = backoff_factor
        self._STATUS_FORCE_LIST = status_forcelist
        self._METHOD_ALLOWLIST = method_allowlist

    def create_query(
        self, params: CreateQueryRunRpcParams
    ) -> CreateQueryRunRpcResponse:
        with self.session.post(
            self.url,
            data=json.dumps(CreateQueryRunRpcRequest(params=[params]).dict()),
            headers=self._headers,
        ) as result:
            data = self._handle_response(result, "createQueryRun")

            create_query_resp = CreateQueryRunRpcResponse(**data)

            return create_query_resp

    def get_query_run(
        self, params: GetQueryRunRpcRequestParams
    ) -> GetQueryRunRpcResponse:
        with self.session.post(
            self.url,
            data=json.dumps(GetQueryRunRpcRequest(params=[params]).dict()),
            headers=self._headers,
        ) as result:

            data = self._handle_response(result, "getQueryRun")

            get_query_run_resp = GetQueryRunRpcResponse(**data)

            return get_query_run_resp

    def get_sql_statement(
        self, params: GetSqlStatementParams
    ) -> GetSqlStatementResponse:
        with self.session.post(
            self.url,
            data=json.dumps(GetSqlStatementRequest(params=[params]).dict()),
            headers=self._headers,
        ) as result:
            data = self._handle_response(result, "getSqlStatement")

            get_sql_statement_resp = GetSqlStatementResponse(**data)

            return get_sql_statement_resp

    def cancel_query_run(
        self, params: CancelQueryRunRpcRequestParams
    ) -> CancelQueryRunRpcResponse:
        with self.session.post(
            self.url,
            data=json.dumps(CancelQueryRunRpcRequest(params=[params]).dict()),
            headers=self._headers,
        ) as result:
            data = self._handle_response(result, "cancelQueryRun")

            cancel_query_run_resp = CancelQueryRunRpcResponse(**data)

            return cancel_query_run_resp

    def get_query_result(
        self, params: GetQueryRunResultsRpcParams
    ) -> GetQueryRunResultsRpcResponse:
        with self.session.post(
            self.url,
            data=json.dumps(GetQueryRunResultsRpcRequest(params=[params]).dict()),
            headers=self._headers,
        ) as result:
            data = self._handle_response(result, "getQueryRunResults")
            get_query_run_results_resp = GetQueryRunResultsRpcResponse(**data)
            return get_query_run_results_resp

    def _handle_response(self, result: requests.Response, method: str) -> dict:
        if result.status_code is None:
            raise ServerError(
                status_code=0,
                message=f"Unable to connect to server when calling `{method}`. Please try again later.",
            )

        if result.status_code >= 500:
            raise ServerError(
                status_code=result.status_code,
                message=f"Unknown server error when calling `{method}`: {result.status_code} - {result.reason}. Please try again later.",
            )

        if result.status_code == 401 or result.status_code == 403:
            raise ApiError("Unauthorized", result.status_code, "Invalid API Key.")

        try:
            data = result.json()
        except json.decoder.JSONDecodeError:
            raise ServerError(
                status_code=result.status_code,
                message=f"Unable to parse response for RPC response from `{method}`: {result.status_code} - {result.reason}. Please try again later.",
            )

        return data

    @property
    def _headers(self) -> dict:
        return {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "x-api-key": self._api_key,
        }

    @property
    def url(self) -> str:
        return f"{self._base_url}/json-rpc"

    @property
    def session(self) -> requests.Session:
        retry_strategy = Retry(
            total=self._MAX_RETRIES,
            backoff_factor=self._BACKOFF_FACTOR,  # type: ignore
            status_forcelist=self._STATUS_FORCE_LIST,
            allowed_methods=self._METHOD_ALLOWLIST,
        )

        adapter = HTTPAdapter(max_retries=retry_strategy)  # type: ignore
        session = requests.Session()
        session.mount("https://", adapter)
        session.mount("http://", adapter)

        return session



from datetime import datetime
from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from .tags import Tags


class QueryRun(BaseModel):
    id: str
    sqlStatementId: str
    state: str
    path: str
    fileCount: Optional[int]
    lastFileNumber: Optional[int]
    fileNames: Optional[str]
    errorName: Optional[str]
    errorMessage: Optional[str]
    errorData: Optional[Any]
    dataSourceQueryId: Optional[str]
    dataSourceSessionId: Optional[str]
    startedAt: Optional[str]
    queryRunningEndedAt: Optional[str]
    queryStreamingEndedAt: Optional[str]
    endedAt: Optional[str]
    rowCount: Optional[int]
    totalSize: Optional[int]
    tags: Tags
    dataSourceId: str
    userId: str
    createdAt: str
    updatedAt: datetime
    archivedAt: Optional[datetime]
